import unittest
from phonepiece.inventory import read_inventory, create_inventory


class TestInventory(unittest.TestCase):

    def test_all(self):
        inv = read_inventory('eng')

        phones = inv.phone.tolist()
        self.assertTrue(phones[0] == '<blk>')
        self.assertTrue(phones[-1] == '<eos>')

        phonemes = inv.phoneme.tolist()
        self.assertTrue(phonemes[0] == '<blk>')
        self.assertTrue(phonemes[-1] == '<eos>')

    def test_allophone(self):

        target_langs = ['eng', 'cmn', 'jpn', 'deu', 'spa', 'fra', 'rus', 'tur', 'ita']

        for lang_id in target_langs:
            inv = read_inventory(lang_id)

            for phone in inv.phone2phoneme.keys():
                self.assertTrue(phone in inv.phone)

            for phoneme in inv.phoneme2phone.keys():
                self.assertTrue(phone in inv.phoneme)


    def test_create_inventory(self):

        inv = create_inventory('eng', ['a', 'b', 'c'])
        self.assertEqual(len(inv.phoneme), 5)
        self.assertEqual(len(inv.phone), 5)


import sys
from pathlib import Path
from phonepiece.distance import phonological_distance, fast_edit_distance, edit_distance
import argparse
from tqdm import tqdm
from collections import Counter

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Read the contents of two files.')
    parser.add_argument( '--hyp', help='the path to the hypothesis file')
    parser.add_argument('--ref', help='the path to the reference file')
    parser.add_argument('-f', '--format', default='kaldi', help='kaldi or text')
    parser.add_argument('-o', '--output', help='the path to the reference file')
    parser.add_argument('-v', '--verbose', type=bool ,default=False)

    args = parser.parse_args()
    output = args.output
    ref = args.ref
    hyp = args.hyp
    verbose = args.verbose
    file_format = args.format

    assert file_format in ['kaldi', 'text']

    w = None
    if output is not None:
        Path(output).parent.mkdir(exist_ok=True, parents=True)
        w = open(output, "w")

    expect_label = {}

    for i, line in enumerate(open(ref, 'r')):
        fields = line.strip().split()

        if file_format == 'kaldi':
            utt_id = fields[0]
            expect_label[utt_id] = fields[1:]
        else:
            utt_id = f"{i:05d}"
            expect_label[utt_id] = fields

    tot_len_cnt = 0
    tot_err_cnt = 0
    tot_dst_cnt = 0
    tot_add_cnt = 0
    tot_sub_cnt = 0
    tot_del_cnt = 0

    add_counter = Counter()
    del_counter = Counter()
    sub_counter = Counter()

    for i, line in tqdm(enumerate(open(hyp, 'r').readlines())):
        fields = line.strip().split()

        if file_format == 'kaldi':
            utt_id = fields[0]
            hyp = fields[1:]
        else:
            utt_id = f"{i:05d}"
            hyp = fields

        if utt_id not in expect_label:
            continue

        ref = expect_label[utt_id]

        dst_cnt = phonological_distance(ref, hyp)
        all_cnt = len(expect_label[utt_id])

        if verbose:
            err_cnt, add_cnt, del_cnt, sub_cnt, add_lst, del_lst, sub_lst = edit_distance(ref, hyp, utt_id=utt_id, verbose=verbose, out=w)
            tot_add_cnt += add_cnt
            tot_del_cnt += del_cnt
            tot_sub_cnt += sub_cnt
            add_counter.update(add_lst)
            del_counter.update(del_lst)
            sub_counter.update(sub_lst)

        else:
            err_cnt = fast_edit_distance(ref, hyp)

        tot_err_cnt += err_cnt
        tot_dst_cnt += dst_cnt
        tot_len_cnt += all_cnt

    print(f"------------------------------------")
    print(f"TOTAL ERR: {tot_err_cnt}\n")
    print(f"TOTAL DST: {tot_dst_cnt}\n")
    print(f"TOTAL LEN: {tot_len_cnt}\n")

    if verbose:
        print(f"TOTAL ADD {tot_add_cnt:.3f}: {add_counter.most_common(10)}")
        print(f"TOTAL DEL {tot_del_cnt:.3f}: {del_counter.most_common(10)}")
        print(f"TOTAL SUB {tot_sub_cnt:.3f}: {sub_counter.most_common(10)}")

    if tot_len_cnt != 0:
        print(f"TOTAL ERR RATE: {tot_err_cnt/tot_len_cnt:.3f}")
        if verbose:
            print(f"TOTAL ADD RATE: {tot_add_cnt/tot_len_cnt:.3f}")
            print(f"TOTAL DEL RATE: {tot_del_cnt/tot_len_cnt:.3f}")
            print(f"TOTAL SUB RATE: {tot_sub_cnt/tot_len_cnt:.3f}")

        print(f"TOTAL DST RATE: {tot_dst_cnt/tot_len_cnt:.3f}")
    print(f"------------------------------------")

    if w is not None:
        w.write(f"------------------------------------")
        w.write(f"TOTAL ERR: {tot_err_cnt}\n")
        w.write(f"TOTAL DST: {tot_dst_cnt}\n")
        w.write(f"TOTAL LEN: {tot_len_cnt}\n")

        if verbose:
            w.write(f"TOTAL ADD {tot_add_cnt:.3f}: {add_counter.most_common(10)})\n")
            w.write(f"TOTAL DEL {tot_del_cnt:.3f}: {del_counter.most_common(10)})\n")
            w.write(f"TOTAL SUB {tot_sub_cnt:.3f}: {sub_counter.most_common(10)})\n")

        if tot_len_cnt != 0:
            w.write(f"TOTAL ERR RATE: {tot_err_cnt/tot_len_cnt:.3f}\n")

            if verbose:
                w.write(f"TOTAL ADD RATE: {tot_add_cnt / tot_len_cnt:.3f}\n")
                w.write(f"TOTAL DEL RATE: {tot_del_cnt / tot_len_cnt:.3f}\n")
                w.write(f"TOTAL SUB RATE: {tot_sub_cnt / tot_len_cnt:.3f}\n")

            w.write(f"TOTAL DST RATE: {tot_dst_cnt/tot_len_cnt:.3f}\n")

        w.write(f"------------------------------------\n")
        w.close()


import numpy as np
from phonepiece.config import PhonePieceConfig
import csv
import unicodedata
from collections import defaultdict

_norm_rules = [
    (':', 'ː'),
    ("ɡ̥", "k"),
    ('g', 'q̠'),
    ('d̥', "t"),
    ("b̥", "b"),
    ("'", "ʼ"),
    ("I", "ɪ"),
    ('O', 'o'),
    ('Y', 'ʏ'),
    ('ɚ', 'ə˞'),
    ('ɝ', 'ɜ˞'),
    ('ә', 'ə'),
    ('ˈ', ''),
    ('tʃ', 't͡ʃ'),
    ('dʒ','d͡ʒ'),
    ('N', 'n'), # special handling for JPN N
    # tones: ˩˨˧˦˥
    ('˩', ''),
    ('˨', ''),
    ('˧', ''),
    ('˦', ''),
    ('˥', ''),
]

# singleton
ipa = None

def read_ipa():
    global ipa

    if ipa is not None:
        return ipa

    feature_file = PhonePieceConfig.data_path / f'ipa_all.csv'

    feature_map = {'0': 0, '-': -1, '+': 1}

    phone2feature = {}

    with open(feature_file, 'r', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)

        # This skips the first row of the CSV file.
        next(reader)

        for row in reader:
            phone = unicodedata.normalize('NFD', row[0])
            features = list(map(lambda x: feature_map[x], row[1:]))
            assert len(features) == 24
            phone2feature[phone] = np.array(features)

    # weights are taken from panphon ipa_weights
    weights = [
       1,1,1,0.5,0.25,0.25,0.25,0.125,0.125,0.125,0.125,0.25,0.25,0.125,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.125,0,0
    ]

    assert len(weights) == 24

    base_phone_file = PhonePieceConfig.data_path / f'ipa_base.csv'
    base_phones = []
    with open(base_phone_file, 'r', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)

        # This skips the first row of the CSV file.
        next(reader)

        for row in reader:
            phone = unicodedata.normalize('NFD', row[0])
            base_phones.append(phone)

    ipa = IPA(phone2feature, weights, base_phones)
    return ipa


class IPA:

    def __init__(self, phone2feature, weights, base_phones):
        self.phone2feature = phone2feature
        self.weights = weights
        self.base_phones = sorted(base_phones)
        self.base_phones_set = set(base_phones)

        self.canonical_phone = {}
        self.compute_canonical_form()

    def __getitem__(self, item):
        item = self.normalize(item)
        if item in self.phone2feature:
            return self.phone2feature[item]
        else:
            return np.zeros(24, dtype=np.int32)

    def __contains__(self, item):
        item = self.normalize(item)
        return item in self.phone2feature


    def read_feature(self, item):
        return self.__getitem__(item)

    def compute_canonical_form(self):
        phone_idx = defaultdict(list)

        power = np.array([2**i for i in range(24)])
        for phone, feat in self.phone2feature.items():
            idx = np.sum((feat + 2)*power)
            phone_idx[idx].append(phone)

        for _, phones in phone_idx.items():
            if len(phones) == 1:
                self.canonical_phone[phones[0]] = phones[0]
            else:
                canonical_form = None

                # identify the canoncial form by
                # checking whether it is in the base phones
                # otherwise pick up the min length phone
                min_length = 100
                min_phone = None

                for phone in phones:
                    if phone in self.base_phones_set:
                        canonical_form = phone
                        break
                    else:
                        if len(phone) < min_length:
                            min_length = len(phone)
                            min_phone = phone


                if canonical_form is None:
                    canonical_form = min_phone

                assert canonical_form is not None
                for phone in phones:
                    self.canonical_phone[phone] = canonical_form

    def normalize(self, orig_phone):
        norm_phone = unicodedata.normalize('NFD', orig_phone)

        # normalize some easy mistakes
        for rule in _norm_rules:
            norm_phone = norm_phone.replace(rule[0], rule[1])

        phone = norm_phone
        # strip diacritics until it find match
        while len(phone) > 0:
            if phone in self.phone2feature:
                return self.canonical_phone[phone]
            phone = phone[:-1]

        # guard cases such as: ʰk
        if len(norm_phone) > 1 and norm_phone[1:] in self.phone2feature:
            return self.canonical_phone[norm_phone[1:]]

        # guard cases such as ʰkʷ
        if len(norm_phone) > 2 and norm_phone[1:-1] in self.phone2feature:
            return self.canonical_phone[norm_phone[1:-1]]

        # give up ... and return empty as an invalid phone to debug
        PhonePieceConfig.logger.error(f"cannot normalize phone {orig_phone}")
        return ''

    def tokenize(self, orig_phone):

        norm_phone = unicodedata.normalize('NFD', orig_phone)

        # normalize some easy mistakes
        for rule in _norm_rules:
            norm_phone = norm_phone.replace(rule[0], rule[1])

        phone = norm_phone

        res = []
        i = 0
        while i < len(phone):
            longest_len = -1

            for j in range(1, min(8, len(phone) - i)+1):
                if phone[i:i+j] in self.phone2feature:
                    longest_len = j

            if longest_len != -1:
                subphone = phone[i:i+longest_len]
                res.append(unicodedata.normalize('NFC', subphone))
                i += len(subphone)
            else:
                i += 1

        return res

    def similarity(self, p1, p2):
        """
        similarity between phone 1 and phone 2

        :param p1:
        :param p2:
        :return:
        """

        p1 = self.normalize(p1)
        p2 = self.normalize(p2)

        # similarity is 0 if either entry is not found
        if p1 not in self.phone2feature or p2 not in self.phone2feature:
            return 0

        f1 = self.phone2feature[p1]
        f2 = self.phone2feature[p2]

        return 1.0 - np.sum(np.abs(f1 - f2) * self.weights) / np.sum(self.weights) / 2.0


    def distance(self, p1, p2):
        return 1.0 - self.similarity(p1, p2)


    def most_similar(self, target_phone, phone_cands, verbose=False):

        max_phone = phone_cands[0]
        min_distance = 1000000

        if target_phone in phone_cands:
            return target_phone

        target_phone = self.normalize(target_phone)

        # if feature not found, then just use the first phone_cand
        if target_phone not in self.phone2feature:
            return phone_cands[0]

        target_feature = self.phone2feature[target_phone]

        for orig_phone in phone_cands:
            phone = self.normalize(orig_phone)

            if phone in self.phone2feature:
                cand_feature = self.phone2feature[phone]
                distance = np.sum(np.abs(target_feature-cand_feature) * self.weights)

                if distance < min_distance:
                    max_phone = orig_phone
                    min_distance = distance

        return max_phone

    def compute_base_phone(self, target_phone):
        return self.most_similar(target_phone, self.base_phones)


import unittest
from phonepiece.epitran import read_epitran_g2p
from phonepiece.inventory import read_inventory

class TestEpitran(unittest.TestCase):

    def test_all(self):
        epi = read_epitran_g2p('spa')
        self.assertTrue(epi['b'] == ['b'])

    def test_g2p_in_phoneme(self):

        langs = ['spa', 'deu', 'fra', 'ita', 'rus', 'tur']

        for lang_id in langs:
            epi = read_epitran_g2p(lang_id)
            inv = read_inventory(lang_id)

            for vs in epi.values():
                for v in vs:
                    self.assertTrue(v in inv.phoneme)


from phonepiece.config import *
from phonepiece.bin.download_model import download_model
from phonepiece.lang import normalize_lang_id
import pip
import importlib

def import_with_auto_install(package, package_name):
    try:
        return importlib.import_module(package)
    except ImportError:
        pip.main(['install', package_name])
    return importlib.import_module(package)


def load_lang_dir(lang_id, model_name='latest'):
    """
    make sure lang_dir is properly downloaded or loaded

    :param lang_id: a 3 char or 2 char iso id
    :param model_name: model name or a customized path
    :type lang_id_or_path:
    :return:
    :rtype:
    """

    # normalize language id (e.g: 2 char 639-1 -> 3 char 639-3)
    lang_id = normalize_lang_id(lang_id)

    # windows does not allow the following directory name, which coincides with language id
    # we give them a _ prefix to avoid naming issue on windows
    if lang_id in ['prn', 'con', 'aux', 'null']:
        lang_id = '_' + lang_id

    # check whether a customized path is used or not
    if (Path(model_name) / lang_id).exists():
        lang_dir = Path(model_name) / lang_id
    else:
        lang_dir = PhonePieceConfig.data_path / 'model' / model_name / lang_id

        # if not exists, we try to download the model
        if not lang_dir.exists():
            download_model(model_name)

        if not lang_dir.exists():
            print(f"warning: could not download or read {lang_id} inventory, using eng instead")
            lang_dir = PhonePieceConfig.data_path / 'model' / model_name / 'eng'

    return lang_dir


from iso639 import languages
import re
import json
from pathlib import Path
from phonepiece.config import PhonePieceConfig


# map macro lang to its individual lang with the largest population or simply the first child entry on Wikipedia
_macro_to_individual = {
    'zho': 'cmn',
    'ara': 'arb',
    'aze': 'azb',
    'fas': 'pes',
    'msa': 'zlm',
    'ori': 'ory',
    'kok': 'gom',
    'srd': 'sro',
    'est': 'ekk',
    'grn': 'gug',
}

def normalize_lang_id(lang_id):

    if len(lang_id) == 3:

        # macro to individual language normalization
        if lang_id in _macro_to_individual:
            return _macro_to_individual[lang_id]

        return lang_id

    assert len(lang_id) == 2

    language = languages.get(part1=lang_id)
    iso3 = language.part3
    if iso3 in _macro_to_individual:
        return _macro_to_individual[iso3]
    return iso3

def read_all_langs():

    tree = read_tree()
    iso_lst = list(tree.iso2path.keys())
    return iso_lst


def read_tree():
    iso2path = {}

    for line in open(PhonePieceConfig.data_path / 'tree.txt', 'r', encoding='utf-8'):
        fields = line.strip().split()
        iso = fields[0]
        path = fields[1].split('/')[1:]

        iso2path[iso] = path

    return LanguageTree(iso2path)


class LanguageTree:

    def __init__(self, iso2path):

        self.iso2path = iso2path

        self.iso_target = []

    def __contains__(self, item):
        return item in self.iso2path

    def setup_target_langs(self, langs):

        self.iso_target = []

        for lang in langs:
            if lang not in self.iso2path:
                print("language: ", lang, " is not a valid lang")
                continue

            self.iso_target.append(lang)

    def similarity(self, lang1, lang2):

        lang1 = normalize_lang_id(lang1)
        lang2 = normalize_lang_id(lang2)

        path1 = self.iso2path[lang1]
        path2 = self.iso2path[lang2]

        common_len = 0
        for i in range(min(len(path1), len(path2))):
            if path1[i] == path2[i]:
                common_len += 1
            else:
                break

        return common_len

    def distance(self, lang1, lang2):

        lang1 = normalize_lang_id(lang1)
        lang2 = normalize_lang_id(lang2)

        common_len = self.similarity(lang1, lang2)

        path1 = self.iso2path[lang1]
        path2 = self.iso2path[lang2]

        return len(path1) + len(path2) - common_len

    def get_nearest_lang(self, iso):

        iso = normalize_lang_id(iso)

        max_score = 1
        max_lang = 'eng'

        for lang in self.iso_target:
            if iso != lang:
                score = self.similarity(iso, lang)
                if score > max_score:
                    max_score = score
                    max_lang = lang

        return max_lang

    def get_nearest_langs(self, iso, num_lang=10):

        iso = normalize_lang_id(iso)

        if iso not in self.iso2path:
            print(f"language {iso} does not exist in db")
            return ['eng']

        score = {}

        for lang in self.iso_target:
            if iso != lang:
                score[lang] = self.similarity(iso, lang)

        lang_ids = [lang_dist[0] for lang_dist in sorted(score.items(), key=lambda x: -x[1])[:num_lang]]
        return lang_ids

    def get_similar_lang2(self, iso, num_lang=10):

        score = {}

        for lang in self.iso_target:
            if iso != lang:
                score[lang] = self.distance(iso, lang)

        return sorted(score.items(), key=lambda x: x[1])[:num_lang]


import unittest
from phonepiece.distance import phonological_distance

class TestDistance(unittest.TestCase):

    def test_all(self):

        self.assertEqual(phonological_distance('a', 'a') == 0.0)
        self.assertEqual(phonological_distance('a', 'a') == 1.0)


from collections import defaultdict
from phonepiece.utils import load_lang_dir
from phonepiece.lang import normalize_lang_id
from phonepiece.inventory import read_inventory
from phonepiece.utils import import_with_auto_install
from phonepiece.arpa import ArpaConverter


def read_lexicon(lang_id, model_name='latest'):

    """
    read lexicon of the lang_id from a prebuilt inventory model or a customized local path

    :param lang_id: a 3 char or 2 char iso id
    :param model_name: model name or a customized path
    :type lang_id_or_path:
    :return:
    :rtype:
    """

    # normalize language id (e.g: 2 char 639-1 -> 3 char 639-3)
    lang_id = normalize_lang_id(lang_id)

    # load or download lang dir
    lang_dir = load_lang_dir(lang_id, model_name)

    inventory = read_inventory(lang_id, model_name)

    lexicon_path = lang_dir / 'lexicon.txt'

    # special handling for English
    if lang_id == 'eng':
        cmudict_module = import_with_auto_install('cmudict', 'cmudict')
        cmudict = cmudict_module.dict()
        converter = ArpaConverter()
        word2phoneme = {}
        for word in cmudict:
            arpa = cmudict[word][0]
            ipas = converter.convert(arpa)
            word2phoneme[word] = ipas

        return Lexicon(lang_id, inventory, word2phoneme)

    elif not lexicon_path.exists():
        # empty lexicon
        return Lexicon(lang_id, inventory, {})

    word2phoneme = {}
    for line in open(str(lexicon_path), 'r', encoding='utf-8'):
        fields = line.strip().split('\t')

        # wrongly formatted lines
        if len(fields) != 2 or len(fields[0]) == 0 or len(fields[1]) == 0:
            continue

        # to lower by default
        word = fields[0].lower()
        phonemes = fields[1].split()
        remapped_phonemes = inventory.remap(phonemes)
        word2phoneme[word] = remapped_phonemes

    return Lexicon(lang_id, inventory, word2phoneme)


def write_lexicon(lexicon, lexicon_path):

    wfile = open(str(lexicon_path), 'w', encoding='utf-8')

    for word, units in lexicon.word_to_unit.items():
        wfile.write(word+' '+' '.join(units)+'\n')

    wfile.close()


class Lexicon:

    def __init__(self, lang_id, inventory, word2phoneme):

        self.lang_id = lang_id

        # word to id
        self.word2phoneme = word2phoneme

        self.inventory = inventory

    def __str__(self):
        return f'<Lexicon: {len(self.word2phoneme)} words>'

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.word2phoneme)

    def __contains__(self, item):
        return item.lower() in self.word2phoneme

    def __getitem__(self, item):
        return self.word2phoneme[item.lower()]


from phonepiece.lang import normalize_lang_id
from phonepiece.config import PhonePieceConfig
from collections import defaultdict
import unicodedata
import csv
from phonepiece.utils import load_lang_dir
import re
from phonepiece.ipa import read_ipa


def read_epitran_g2p(lang_id_or_g2p_id, writing_system=None):

    if '-' in lang_id_or_g2p_id:
        fields = lang_id_or_g2p_id.split('-')
        assert len(fields) == 2

        lang_id = fields[0]
        writing_system = fields[1]
    else:
        lang_id = normalize_lang_id(lang_id_or_g2p_id)

    lang_dir = load_lang_dir(lang_id)
    g2p_dir = lang_dir / 'g2p'

    if writing_system is not None:
        g2p_path = g2p_dir / f"{lang_id}-{writing_system}.csv"
        assert g2p_path.exists(), f"{lang_id}-{writing_system} does not exist!"

    else:
        if g2p_dir.exists():
            g2p_ids = sorted([p.stem for p in list(g2p_dir.glob('*'))])
        else:
            g2p_ids = []

        assert len(g2p_ids) > 0, "provided lang_id " + lang_id + " is not supported by epitran"

        if len(g2p_ids) > 1:
            print(f"there are multiple g2p_ids {g2p_ids}, we use the first {g2p_ids[0]}")

        g2p_id = g2p_ids[0]
        g2p_path = g2p_dir / (g2p_id+".csv")

    g2p = defaultdict(list)

    ipa = read_ipa()

    with open(g2p_path, encoding='utf-8') as f:
        reader = csv.reader(f)
        orth, phon = next(reader)
        if orth != 'Orth' or phon != 'Phon':
            raise DatafileError(f'Header is ["{orth}", "{phon}"] instead of ["Orth", "Phon"].')
        for (i, fields) in enumerate(reader):
            try:
                graph, phons = fields
            except ValueError as malformed_data_file:
                raise DatafileError(f'Map file is not well formed at line {i + 2}.') from malformed_data_file

            phons = ipa.tokenize(phons)
            for phon in phons:
                graph = unicodedata.normalize('NFC', graph)
                phon = unicodedata.normalize('NFC', phon)
                phon = re.sub('[˩˨˧˦˥]', '', phon)
                g2p[graph].append(phon)

    return g2p


# -*- coding: utf-8 -*-
# this file is adapted from epitran

import csv
from phonepiece.config import PhonePieceConfig
from phonepiece.ipa import read_ipa

def none2str(x):
    return x if x else ''


class ArpaConverter:

    def __init__(self):
        arpa_path = PhonePieceConfig.data_path / 'arpabet.csv'

        self.ipa = read_ipa()

        self.arpa_map = {}
        self.ipa_to_arpa = {}

        self.ipa_set = set()

        with open(arpa_path, 'r', encoding='utf-8') as f:
            reader = csv.reader(f)
            for arpa, ipa in reader:
                ipa = ipa.strip()
                self.arpa_map[arpa] = ipa.split()

                # only register 1 ipa 1 arpa mapping
                if len(ipa.split()) > 1:
                    continue

                self.ipa_to_arpa[ipa] = arpa

                self.ipa_set.update(ipa.split())

    def convert(self, str_or_lst):
        if isinstance(str_or_lst, str):
            str_or_lst = str_or_lst.split(' ')

        result = []
        for arpa in str_or_lst:
            if str.isdigit(arpa[-1]):
                arpa = arpa[:-1]

            result.extend(self.arpa_map[arpa.lower()])

        return result

    def convert_ipa_to_arpa(self, ipa_lst):

        normalized_ipa = []
        for ipa in ipa_lst:
            normalized_ipa.append(self.ipa.most_similar(ipa, sorted(list(self.ipa_set))))

        arpa_lst = []

        while len(normalized_ipa) > 0:
            two_ipa_arpa = " ".join(normalized_ipa[:2])
            if two_ipa_arpa in self.ipa_to_arpa:
                arpa_lst.append(self.ipa_to_arpa[two_ipa_arpa])
                normalized_ipa = normalized_ipa[2:]
                continue

            ipa = normalized_ipa[0]
            if ipa in self.ipa_to_arpa:
                arpa_lst.append(self.ipa_to_arpa[ipa])

            normalized_ipa = normalized_ipa[1:]

        return arpa_lst


import unittest
from phonepiece.ipa import read_ipa


class TestIPA(unittest.TestCase):

    def test_all(self):
        ipa = read_ipa()

        # normalize
        self.assertEqual(ipa.normalize('a') , 'a')
        self.assertEqual(ipa.normalize('a:') , 'aː')
        self.assertEqual(ipa.normalize('b') , 'b')
        self.assertEqual(ipa.normalize('g') , 'q')
        self.assertEqual(ipa.normalize('d̠') , 'd')

        # score
        self.assertEqual(ipa.similarity('a', 'a') , 1.0)
        self.assertEqual(ipa.similarity('g', 'q̠') , 1.0)
        self.assertTrue(ipa.similarity('a', 'e') > ipa.similarity('a', 'b'))
        self.assertTrue(ipa.similarity('a', 'ɑ') > ipa.similarity('a', 'i'))
        self.assertTrue(ipa.similarity('a', 'ɑ') > ipa.similarity('ɑ', 'i'))

        # most similar
        self.assertEqual(ipa.most_similar('a', ['b', 's', 'r', 'e']) , 'e')
        self.assertEqual(ipa.most_similar('b', ['p', 's', 'a', 'e']) , 'p')
        self.assertEqual(ipa.most_similar('b', ['b', 's', 'a', 'e']) , 'b')

        self.assertEqual(ipa.tokenize('kʰæt'), ['kʰ', 'æ', 't'])
        self.assertEqual(ipa.tokenize('kʰæ t'), ['kʰ', 'æ', 't'])
        self.assertEqual(ipa.tokenize('a') , ['a'])
        self.assertEqual(ipa.tokenize('aa') , ['a', 'a'])
        self.assertEqual(ipa.tokenize('ab') , ['a', 'b'])
        self.assertEqual(ipa.tokenize('a  a') , ['a', 'a'])

        # remove tones symbols
        self.assertEqual(ipa.tokenize('ʔɓaːn˧˧') , ['ʔ', 'ɓ', 'aː', 'n'])

        # normalizing into the base forms
        self.assertEqual(ipa.compute_base_phone('a'), 'a')
        self.assertEqual(ipa.compute_base_phone('tː'), 't')

